/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    armarx_control::ArmarXObjects::NJointControllerExample
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#include "Component.h"

#include <chrono>
#include <cmath>
#include <filesystem>
#include <thread>

#include <Eigen/Geometry>

#include "ArmarXCore/core/time/Metronome.h"
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/Clock.h>
#include <ArmarXCore/core/time/ClockType.h>
#include <ArmarXCore/core/time/forward_declarations.h>
#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>


namespace armarx::control::components::example_client
{

    Component::Component()
    {
        addPlugin(virtualRobotReaderPlugin);
    }


    Component::~Component() = default;

    void
    Component::onInitComponent()
    {
        // pass
    }


    void
    Component::onConnectComponent()
    {
        robot = virtualRobotReaderPlugin->get().getRobot(properties.robotName);

        ARMARX_CHECK_NOT_NULL(robot);
        ARMARX_CHECK(virtualRobotReaderPlugin->get().synchronizeRobot(*robot, Clock::Now()));

        task = new armarx::RunningTask<Component>(this, &Component::runExample);
        task->start();
    }


    void
    Component::onDisconnectComponent()
    {
        const bool join = true;
        task->stop(join);
    }


    void
    Component::onExitComponent()
    {
        // pass
    }


    std::string
    Component::getDefaultName() const
    {
        return "NJointControllerExample";
    }

    void
    Component::runExample()
    {
        auto robotUnit = getControlComponentPlugin().getRobotUnitPlugin().getRobotUnit();

        // Load the library that our controller is defined in.
        robotUnit->loadLibFromPackage("armarx_control", "libarmarx_control_njoint_controller.so");

        // We use a ControllerBuilder to configure the controller. This controller builder is a convenient and safe (!)
        // way to create a NJointController.

        // There are various options of how to set the configuration:
        // (1) load a pre-defined configuration file
        // (2) set the configuration programatically

        auto builder = controllerBuilder<ControllerType>();
        builder.withNodeSet(properties.rns);

        // load pre-defined config from file
        {
            const std::string controllerClassName =
                control::common::ControllerTypeNames.to_name(ControllerType);

            const armarx::PackagePath configPath("armarx_control",
                                                 "controller_config/" + controllerClassName + "/" +
                                                     properties.configName + ".json");

            ARMARX_IMPORTANT << "Using config `" << configPath.toSystemPath() << "`.";
            ARMARX_CHECK(std::filesystem::exists(configPath.toSystemPath()));
            builder.withConfig(configPath.toSystemPath());

            ARMARX_INFO << VAROUT(builder.config().kpImpedance);
            ARMARX_INFO << VAROUT(builder.config().kdImpedance);
            ARMARX_INFO << VAROUT(builder.config().kpNullspace);
            ARMARX_INFO << VAROUT(builder.config().kdNullspace);
            ARMARX_INFO << VAROUT(builder.config().desiredPose);
            ARMARX_INFO << VAROUT(builder.config().desiredTwist);
            if (builder.config().desiredNullspaceJointAngles.has_value())
                ARMARX_INFO << VAROUT(builder.config().desiredNullspaceJointAngles.value());
            ARMARX_INFO << VAROUT(builder.config().torqueLimit);
            ARMARX_INFO << VAROUT(builder.config().qvelFilter);
        }

        ARMARX_CHECK_NOT_NULL(robot);
        ARMARX_CHECK(virtualRobotReaderPlugin->get().synchronizeRobot(*robot, armarx::Clock::Now()));
        /// !!!!
        /// you have to wait for enough time to let the clone finish. Otherwise, you may get random / wrong values below when you read the joint angles and TCP pose, etc.
        /// !!!!
        Clock::WaitFor(Duration::MilliSeconds(1000));

        const auto rns = robot->getRobotNodeSet(properties.rns);

        // Set configuration programmatically. Use this for those parts of the config that cannot be specified in advance.
//        builder.config().desiredNullspaceJointAngles = rns->getJointValuesEigen();
        ARMARX_INFO << VAROUT(rns->getJointValuesEigen());
        builder.config().desiredPose = rns->getTCP()->getPoseInRootFrame();
        ARMARX_INFO << VAROUT(builder.config().desiredPose);

        ARMARX_TRACE;
        ARMARX_INFO << "Creating controller";
        auto ctrlWrapper = builder.create();

        // The controller is scoped by default, which means that if the ControllerWrapper goes out of scope, it will deactivate
        // and delete the NJointController. To prevent this, you can use the daemonize() method if needed.
        ctrlWrapper->daemonize(true);

        ARMARX_TRACE;
        ARMARX_CHECK(ctrlWrapper.has_value());
        ctrl.emplace(std::move(ctrlWrapper.value()));

        Clock::WaitFor(Duration::MilliSeconds(300));
        ARMARX_INFO << "Activating controller";
        ctrl->updateConfig();
        ctrl->activate();

        // Now the controller is active. In our example, the arm should now be compliant.
        // To move the arm, we can update the config as shown below. In this example,
        // the robot will execute a circular motion.

        const Eigen::Isometry3f robot_T_tcp_initial(rns->getTCP()->getPoseInRootFrame());

        Metronome metronome(Frequency::Hertz(20));
        ARMARX_INFO << "Starting movement";

        const auto start = Clock::Now();
        while (Clock::Now() - start < Duration::Seconds(properties.movementDurationSeconds))
        {
            const float t = (Clock::Now() - start).toSecondsDouble();
            const float T = properties.movementDurationSeconds;
            const float x = 2 * M_PIf32 * t / T;

            const Eigen::Vector3f posOffset{
                properties.movementRadius * std::sin(x), std::cos(x) - 1, 0};

            const Eigen::Isometry3f global_T_tcp_desired =
                Eigen::Translation3f(posOffset) * robot_T_tcp_initial;

//            // update config
//            ctrl->config.desiredPose = global_T_tcp_desired.matrix();

//            // send new config to controller
//            ctrl->updateConfig();

            metronome.waitForNextTick();
        }

        ctrl->deactivate();
    }


    armarx::PropertyDefinitionsPtr
    Component::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new armarx::ComponentPropertyDefinitions(getConfigIdentifier());

        def->optional(properties.robotName, "robotName");
        def->optional(properties.configName, "configName");
        def->optional(properties.movementRadius, "movementRadius");
        def->optional(properties.movementDurationSeconds, "movementDurationSeconds");


        return def;
    }


    ARMARX_DECOUPLED_REGISTER_COMPONENT(Component);

} // namespace armarx::navigation::components::example_client
