/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once


#include <cmath>
#include <cstdint>
#include <limits>
#include <type_traits>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <armarx/control/hardware_config/Types.h>

namespace armarx::control::ethercat
{

    template <typename T, typename SignedT = std::int64_t>
    class ModularConvertedValue
    {

    public:
        ModularConvertedValue()
        {
            rawValue = nullptr;
            zeroOffset = 0;
            discontinuityOffset = 0;
            maxValue = 0;
            factor = 0;
            isInverted = false;

            static_assert(std::is_signed_v<SignedT>);

            if constexpr(std::is_unsigned_v<T>)
            {
                static_assert(sizeof(SignedT) > sizeof(T));
            }

            if constexpr(std::is_signed_v<T>)
            {
                static_assert(sizeof(SignedT) >= sizeof(T));
            }
        }

        void
        init(T* raw,
             const hardware_config::types::ModularConvertedValueConfig modularConfig,
             const char* nameForDebugging = "")
        {
            init(raw,
                 modularConfig.getZeroOffset(),
                 modularConfig.getDiscontinuityOffset(),
                 modularConfig.getMaxValue(),
                 modularConfig.getIsInverted(),
                 nameForDebugging);
        }

        void
        init(T* raw,
             float zeroOffset,
             float discontinuityOffset,
             float maxValue,
             bool isInverted,
             const char* nameForDebugging = "")
        {
#if 0
            const auto rawAsInt = reinterpret_cast<std::uint64_t>(raw);
            ARMARX_CHECK_EQUAL((rawAsInt % alignof(T)), 0)
                << "\nThe alignment is wrong!"
                << "\nIt has to be " << alignof(T) << ", but the data is aligned with "
                << rawAsInt % alignof(std::max_align_t) << "!"
                << "\nThis is an offset of " << (rawAsInt % alignof(T)) << " bytes!"
                << "\nThe datatype is " << GetTypeString<T>() << "\nIts size is " << sizeof(T)
                << "\nraw = " << raw << " bytes"
                << "\nThe name is '" << nameForDebugging << "'";
#endif
            ARMARX_CHECK_POSITIVE(maxValue)
                << "max value must not be zero (cannot divide/modulo by 0).";

            ARMARX_CHECK_LESS_EQUAL(maxValue, std::numeric_limits<T>::max())
                << "max value is greater than the maximum value of the datatype, "
                << "the datatype is " << GetTypeString<T>();

            this->rawValue = raw;
            this->zeroOffset = static_cast<SignedT>(zeroOffset);
            this->discontinuityOffset = static_cast<SignedT>(discontinuityOffset);
            this->maxValue = static_cast<SignedT>(maxValue);
            this->factor = 2 * M_PIf64 / maxValue;
            this->isInverted = isInverted;

            read();
        }

        void
        read()
        {
            // "%" operator is not the real mathematical modulo operation but some creative ISO
            // reinterpretation. This fixes it.
            auto actual_real_modulo = [](SignedT dividend, SignedT divisor) -> SignedT
            {
                const SignedT result = dividend % divisor;
                return result >= 0 ? result : result + divisor;
            };

            const SignedT rawValueSigned = static_cast<SignedT>(*rawValue);
            const SignedT maxValueSigned = static_cast<SignedT>(maxValue);


            // Normalize direction of rotation.
            const SignedT sign = isInverted ? -1 : 1;
            const SignedT dirnorm = actual_real_modulo((sign * rawValueSigned), maxValueSigned);

            // Raw values within monotonic_raw_value are guaranteed to not fall into a
            // discontinuity.
            const SignedT monotonic_dirnorm =
                actual_real_modulo((dirnorm + discontinuityOffset), maxValueSigned);

            // Zero shift.
            const SignedT zeroshift_monotonic_dirnorm = monotonic_dirnorm - zeroOffset;

            // Convert ticks to rad.
            const double zeroshift_monotonic_dirnorm_rad =
                static_cast<double>(zeroshift_monotonic_dirnorm) * factor;

            value = zeroshift_monotonic_dirnorm_rad;
        }

        void
        write()
        {
            //*rawValue = static_cast<T>((value / factor) - offset);
        }

        double value;

        T
        getRaw() const
        {
            return *rawValue;
        }

    private:
        T* rawValue;
        SignedT zeroOffset;
        SignedT discontinuityOffset;
        T maxValue;
        double factor;
        bool isInverted;
    };

} // namespace armarx::control::ethercat
