#include "Pointing.h"

#include <ArmarXCore/core/time.h>

#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>

#include <armarx/control/pointing/constants/constants.h>

#include "PointingIK.h"

namespace armarx::control::pointing::core
{
    Pointing::Pointing(const Remote& remote) : remote_(remote)
    {
    }

    void
    Pointing::execute(const Pointing::Parameters params)
    {
        aborted = false;

        ARMARX_INFO << "Pointing with " << ((params.side == Side::LEFT) ? "left" : "right")
                    << " arm at " << params.framedTarget << " in frame '" << params.frame << "'"
                    << (params.handShape ? " with hand shape '" + *params.handShape + "'" : "");

        remote_.trajectoryPlayer->stopTrajectoryPlayer();

        VirtualRobot::RobotPtr robot = getRobot();

        auto target = PositionInGlobalFrame(robot, params.frame, params.framedTarget);
        visualizeTarget(target);

        armarx::TrajectoryPtr traj = computTrajectory(robot, params.side, target);

        if (params.handShape)
        {
            auto shapeJointValues = GetShapeJointValues(robot, params.side, *params.handShape);
            addHandTrajectory(robot, traj, shapeJointValues);
        }

        traj = traj->calculateTimeOptimalTrajectory(
            params.maxVelocity, params.maxAcceleration, 0.0, IceUtil::Time::milliSeconds(10));

        playTrajectory(traj);
    }

    void
    Pointing::abort()
    {
        aborted = true;
    }

    VirtualRobot::RobotPtr
    Pointing::getRobot()
    {
        armarx::armem::robot_state::VirtualRobotReader robotReader;
        robotReader.connect(remote_.memoryNameSystem);

        auto robot = robotReader.getSynchronizedRobot(remote_.robotName);
        if (not robot)
        {
            ARMARX_ERROR << "Could not get synchronized robot '" << remote_.robotName << "'";
            throw PointingFailedException();
        }
        return robot;
    }

    void
    Pointing::visualizeTarget(const Eigen::Vector3f& target)
    {
        auto layer = remote_.arviz.layer(constants::ARVIZ_LAYER_NAME);
        layer.add(armarx::viz::Sphere("Pointing target")
                      .radius(50)
                      .color(armarx::viz::Color::orange())
                      .position(target));
        remote_.arviz.commit(layer);
    }

    TrajectoryPtr
    Pointing::computTrajectory(VirtualRobot::RobotPtr robot, Side side, Eigen::Vector3f target)
    {
        ARMARX_VERBOSE << "Computing trajectory...";
        auto ik = PointingIK(robot, side);
        armarx::TrajectoryPtr traj = ik.computeTrajectory(target);
        if (not traj)
        {
            ARMARX_WARNING << "IK didn't find a solution";
            throw PointingFailedException();
        }
        ARMARX_VERBOSE << "Done computing trajectory";

        if (traj->getLength() == 1) // already pointing at target
        {
            return armarx::TrajectoryPtr(new Trajectory);
        }
        return traj;
    }

    void
    Pointing::addHandTrajectory(VirtualRobot::RobotPtr robot,
                                armarx::TrajectoryPtr& traj,
                                const std::map<std::string, float>& shapeJointValues)
    {
        for (const auto& [jointName, targetValue] : shapeJointValues)
        {
            traj->addDimension({robot->getRobotNode(jointName)->getJointValue(), targetValue},
                               Ice::DoubleSeq(),
                               jointName);
        }
    }

    void
    Pointing::playTrajectory(armarx::TrajectoryPtr traj)
    {
        ARMARX_INFO << "Playing trajectory (" << traj->getTimeLength() << "s)";
        remote_.trajectoryPlayer->loadJointTraj(traj);
        remote_.trajectoryPlayer->startTrajectoryPlayer();


        using namespace armarx::core::time;
        DateTime waitUnil = DateTime::Now() + Duration::Seconds(traj->getTimeLength());

        Clock clock;
        while (DateTime::Now() < waitUnil)
        {
            if (aborted)
            {
                ARMARX_INFO << "Pointing aborted. Stopping trajectory";
                remote_.trajectoryPlayer->stopTrajectoryPlayer();
                throw PointingAbortedException();
            }
            clock.waitFor(Duration::MilliSeconds(10));
        }
        ARMARX_INFO << "Pointing target reached";
    }

    std::map<std::string, float>
    Pointing::GetShapeJointValues(VirtualRobot::RobotPtr robot,
                                  Side side,
                                  const std::string& shapeName)
    {
        auto eefName = std::string("Hand_") + (side == Side::LEFT ? "L" : "R") + "_EEF"; // TODO
        auto eef = robot->getEndEffector(eefName);

        auto shape = eef->getPreshape(shapeName);
        if (not shape)
        {
            ARMARX_ERROR << "Hand '" << eefName << "' has no shape '" << shape << "'";
            throw PointingFailedException();
        }
        return shape->getRobotNodeJointValueMap();
    }

    Eigen::Vector3f
    Pointing::PositionInGlobalFrame(VirtualRobot::RobotPtr robot,
                                    const std::string& frame,
                                    const Eigen::Vector3f& framedPosition)
    {
        auto node = robot->getRobotNode(frame);
        if (not node)
        {
            ARMARX_ERROR << "Robot '" << robot->getName() << "' has no node '" << frame << "'";
            throw PointingFailedException();
        }
        return node->toGlobalCoordinateSystemVec(framedPosition);
    }

} // namespace armarx::control::pointing::core
