#pragma once

#include <VirtualRobot/VirtualRobot.h>

#include <RobotAPI/components/ArViz/Client/Client.h>
#include <RobotAPI/interface/components/TrajectoryPlayerInterface.h>
#include <RobotAPI/libraries/armem/client/MemoryNameSystem.h>
#include <RobotAPI/libraries/core/Trajectory.h>

#include "Side.h"

namespace armarx::control::pointing::core
{

    /**
     * @brief Implementation of a 'Pointing'-gesture.
     *
     * Lets the robot point somewhere by aligning the forearm with the passed target.
     *
     * @ingroup Library-pointing
     */
    class Pointing
    {
    public:
        struct Remote
        {
            std::string robotName;
            armarx::viz::Client arviz;
            armarx::armem::client::MemoryNameSystem memoryNameSystem;
            armarx::TrajectoryPlayerInterfacePrx trajectoryPlayer;
        };

        struct Parameters
        {
            /** Side of the arm to point with. */
            Side side;

            /** Name of the robot node relative to whose frame @ref framedTarget is given. */
            std::string frame;

            /** Position to point at in frame @ref frame. */
            Eigen::Vector3f framedTarget;

            /** Name of the hand preshape to use. If none is provided, the hand will be ignored. */
            std::optional<std::string> handShape;

            /** Maximum velocity */
            double maxVelocity;

            /** Maximum acceleration */
            double maxAcceleration;
        };

        class PointingException : public std::exception
        {
        };

        class PointingFailedException : public PointingException
        {
        };

        class PointingAbortedException : public PointingException
        {
        };


    public:
        Pointing(const Remote& remote);

        /**
         * @brief Executes the pointing gesture.
         *
         * Let's the robot point at the passed target. Blocks unitl the target configuration is
         * reached or the execution is aborted with @ref abort.
         *
         * @param params Parameterization of the pointing gesture.
         *
         * @throws PointingFailedException Thrown if the pointing gesture could not be executed.
         * @throws Thrown when the execution is aborted with @ref abort before reaching the target
         *         configuration.
         */
        void execute(const Parameters params);

        /**
         * @brief Aborts a running execution.
         */
        void abort();


    private:
        VirtualRobot::RobotPtr getRobot();

        void visualizeTarget(const Eigen::Vector3f& target);

        armarx::TrajectoryPtr
        computTrajectory(VirtualRobot::RobotPtr robot, Side side, Eigen::Vector3f target);

        void addHandTrajectory(VirtualRobot::RobotPtr robot,
                               armarx::TrajectoryPtr& traj,
                               const std::map<std::string, float>& shapeJointValues);

        void playTrajectory(armarx::TrajectoryPtr traj);

        static Eigen::Vector3f PositionInGlobalFrame(VirtualRobot::RobotPtr robot,
                                                     const std::string& frame,
                                                     const Eigen::Vector3f& framedTarget);

        static std::map<std::string, float>
        GetShapeJointValues(VirtualRobot::RobotPtr robot, Side side, const std::string& shapeName);


    private:
        Remote remote_;
        std::atomic_bool aborted;
    };

} // namespace armarx::control::pointing::core
