#include "PointingIK.h"

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/RobotNodeSet.h>

namespace armarx::control::pointing::core
{

    PointingIK::PointingIK(VirtualRobot::RobotPtr robot, Side side)
    {
        auto humanMapping = robot->getHumanMapping();
        ARMARX_CHECK(humanMapping) << "Robot '" << robot->getName() << "' has no human mapping";

        auto armDescription = (side == LEFT) ? humanMapping->leftArm : humanMapping->rightArm;
        auto elbow = armDescription.segments.elbow.nodeName;
        auto wrist = armDescription.segments.wrist.nodeName;
        auto palm = armDescription.segments.palm.nodeName;
        auto tcp = armDescription.segments.tcp.nodeName;

        // actuated joints = all arm joints until (excluding) wrist
        auto nodeSetArm = robot->getRobotNodeSet(armDescription.nodeSet);
        auto actuatedJoints = nodeSetArm->getNodeNames();
        actuatedJoints.resize(nodeSetArm->getRobotNodeIndex(wrist));

        // reduced robot must contain shoulder, elbow, wrist and palm for IK to work
        reducedRobot_ = std::make_unique<simox::control::simox::robot::Robot>(
            robot, actuatedJoints, std::vector({wrist, palm, tcp}), true);

        // use elbow - tcp (instead of wrist) as forearm to line up with the target
        ik_ = std::make_unique<simox::control::method::example::PointingIK>(
            *reducedRobot_,
            *reducedRobot_->getNode(tcp),
            *reducedRobot_->getNode(elbow),
            side == LEFT ? reducedRobot_->humanMapping()->leftArm
                         : reducedRobot_->humanMapping()->rightArm);
    }

    TrajectoryPtr
    PointingIK::computeTrajectory(Eigen::Vector3f& target)
    {
        ik_->setPointingTarget(target.cast<double>() / 1000); // simox uses meters

        // actuated joints x timesteps
        std::vector<std::vector<double>> configs(ik_->getNodeNames().size());
        auto appendConfig = [&configs](Eigen::VectorXd config)
        {
            for (size_t i = 0; i < configs.size(); i++)
            {
                configs[i].push_back(config[i]);
            }
        };

        appendConfig(ik_->getJointValues());

        Eigen::VectorXd prevJointValues = ik_->getJointValues();
        while (not ik_->isTargetReached())
        {
            ik_->optimize();
            if ((ik_->getJointValues() - prevJointValues).norm() < 1e-6)
            {
                return nullptr;
            }

            appendConfig(ik_->getJointValues());
            prevJointValues = ik_->getJointValues();
        }

        return new Trajectory(configs, Ice::DoubleSeq(), ik_->getNodeNames());
    }

} // namespace armarx::control::pointing::core
