#pragma once

#include <RobotAPI/libraries/core/Trajectory.h>

#include "Side.h"
#include <simox/control/impl/simox/robot/Robot.h>
#include <simox/example/method/PointingIK.h>

namespace armarx::control::pointing::core
{

    /**
     * @brief IK to compute joint configurations for a pointing gesture.
     *
     * Wrapper for simox-control's method::example::PointingIK. Computes pointing gesture joint
     * configurations that line up the forearm with the target. Uses the robot's human mapping to
     * obtain relevant joints: All arm-joints until (exluding) the wrist are beeing used.
     *
     * @ingroup Library-pointing
     */
    class PointingIK
    {
    public:
        /**
         * @brief Initialize a new PointingIK.
         * @param robot Robot model with current robot configuration to be used by the IK. Must
         *        provide a human mapping.
         * @param side Side of the arm to point with.
         */
        PointingIK(VirtualRobot::RobotPtr robot, Side side);


        /**
         * @brief Compute a joint trajectory ending in a configuration in which the robot points in
         *        the direction of the passed target position.
         *
         * The initial timestep of the trajectory contains the current joint configuration of the
         * robot passed to the constructor. The following timesteps contain the step-by-step
         * optimized configurations until the end configuration is reached. The trajectory is
         * constructed without timestamps, i.e., has a duration of 1s with equidistant timesteps.
         *
         * Neither the generated trajectory nor the end configuration are guaranteed to be self-
         * collision-free. Furthermore, generated trajectories aren't efficient in reaching the end
         * configuration.
         *
         * This method may only be called once to ensure a well-defined initial robot configuration.
         *
         * @param target Target position to point at in the global coordinate system (mm).
         * @return The computed trajectory. Null if no solution was found. If the robot has already
         *         pointed at the target, the trajectory only contains the initial timestep.
         */
        TrajectoryPtr computeTrajectory(Eigen::Vector3f& target);

    private:
        std::unique_ptr<simox::control::simox::robot::Robot> reducedRobot_;
        std::unique_ptr<simox::control::method::example::PointingIK> ik_;
    };

} // namespace armarx::control::pointing::core
